import torch.optim as optim
import torch.nn as nn
import torch.nn.functional as F
from copy import deepcopy
import torch
from tqdm import tqdm

from ...utils.utils import *
from ...utils.metrics import get_loss

class BaseModel(nn.Module):
    def __init__(self, device = 'cpu'):
        super(BaseModel, self).__init__()
        self.device = device

    def fit(self, 
            train_input, 
            train_target, 
            train_states=None, 
            train_graph=None, 
            train_dynamic_graph=None,
            val_input=None, 
            val_target=None,
            val_states=None, 
            val_graph= None, 
            val_dynamic_graph=None,
            loss='mse', 
            epochs=1000, 
            batch_size=10,
            lr=1e-3, 
            initialize=True, 
            verbose=False, 
            patience=100, 
            **kwargs):
        if initialize:
            self.initialize()
        
        optimizer = torch.optim.Adam(self.parameters(), lr=lr)
        loss_fn = get_loss(loss)

        training_losses = []
        validation_losses = []
        early_stopping = patience
        best_val = float('inf')
        for epoch in tqdm(range(epochs)):
            loss = self.train_epoch(optimizer = optimizer, loss_fn = loss_fn, feature = train_input,  target = train_target, batch_size = batch_size, device = self.device)
            training_losses.append(loss)

            val_loss, output = self.evaluate(loss_fn = loss_fn, feature = val_input,  target = val_target, device = self.device)
            validation_losses.append(val_loss)

            if best_val > val_loss:
                best_val = val_loss
                self.output = output
                best_weights = deepcopy(self.state_dict())
                patience = early_stopping
            else:
                patience -= 1

            if epoch > early_stopping and patience <= 0:
                break

            if verbose and epoch%50 == 0:
                print(f"######### epoch:{epoch}")
                print("Training loss: {}".format(training_losses[-1]))
                print("Validation loss: {}".format(validation_losses[-1]))

        print("\n")
        print("Final Training loss: {}".format(training_losses[-1]))
        print("Final Validation loss: {}".format(validation_losses[-1]))

        self.load_state_dict(best_weights)

        
    def train_epoch(self, optimizer, loss_fn, feature, target = None, batch_size = 1, device = 'cpu'):
        """
        Trains one epoch with the given data.
        :param feature: Training features of shape (num_samples, num_nodes,
        num_timesteps_train, num_features).
        :param target: Training targets of shape (num_samples, num_nodes,
        num_timesteps_predict).
        :param batch_size: Batch size to use during training.
        :return: Average loss for this epoch.
        """
        permutation = torch.randperm(feature.shape[0])

        epoch_training_losses = []
        for i in range(0, feature.shape[0], batch_size):
            self.train()
            optimizer.zero_grad()

            indices = permutation[i:i + batch_size]
            X_batch, y_batch = feature[indices], target[indices]
            X_batch = X_batch.to(device=device)
            y_batch = y_batch.to(device=device)

            out = self.forward(X_batch)
            loss = loss_fn(out, y_batch)
            loss.backward()
            optimizer.step()
            epoch_training_losses.append(loss.detach().cpu().numpy())
        return sum(epoch_training_losses)/len(epoch_training_losses)
    
    def evaluate(self, loss_fn, feature, target = None, device = 'cpu'):
        with torch.no_grad():
            self.eval()
            feature = feature.to(device=device)
            target = target.to(device=device)

            out = self.forward(feature)
            val_loss = loss_fn(out, target).to(device="cpu")
            val_loss = val_loss.detach().numpy().item()
            
            return val_loss, out

    def predict(self, feature, graph=None, states=None, dynamic_graph=None):
        """
        Returns
        -------
        torch.FloatTensor
        """
        self.eval()

        return self.forward(feature)



import torch
import torch.nn as nn
import numpy as np
from .utils import *
import os
from tqdm import tqdm
from fastdtw import fastdtw

class Compose:
    """Composes several transforms together. This transform does not support torchscript.
    Please, see the note below.

    Args:
        transforms (list of ``Transform`` objects): list of transforms to compose.

    Example:
        >>> transforms.Compose([
        >>>     transforms.CenterCrop(10),
        >>>     transforms.ToTensor(),
        >>> ])

    .. note::
        In order to script the transformations, please use ``torch.nn.Sequential`` as below.

        >>> transforms = torch.nn.Sequential(
        >>>     transforms.CenterCrop(10),
        >>>     transforms.Normalize((0.485, 0.456, 0.406), (0.229, 0.224, 0.225)),
        >>> )
        >>> scripted_transforms = torch.jit.script(transforms)

        Make sure to use only scriptable transformations, i.e. that work with ``torch.Tensor``, does not require
        `lambda` functions or ``PIL.Image``.

    """

    def __init__(self, transforms, device="cpu"):
        self.transforms = transforms
        self.device = device
        self.feat_mean = 0
        self.feat_std = 1

    def __call__(self, data):
        for key, d in data.items():
            if data[key] is not None:
                data[key] = torch.Tensor(data[key])
                
        for dt, ts in self.transforms.items():
            for t in ts:
                t = t.to(self.device)
                data[dt] = t(data[dt], device=self.device)
                if type(t).__name__ == 'normalize_feat':
                    self.feat_mean, self.feat_std = t.means, t.stds
        return data

    def __repr__(self):
        format_string = self.__class__.__name__ + '('
        for t in self.transforms:
            format_string += '\n'
            format_string += '    {0}'.format(t)
        format_string += '\n)'
        return format_string

class normalize_feat(nn.Module):
    """
    A normalization module for feature standardization in PyTorch. This module adjusts features
    to have zero mean and unit variance along specified dimensions, handling 3D and 4D tensors.

    Parameters
    ----------
    dim : int, optional
        The dimension over which to calculate the mean and standard deviation for normalization. Default: 1.
    """
    def __init__(self, dim=1):
        super().__init__()
        self.dim = dim
        self.means = None
        self.stds = None
    
    def forward(self, X, device='cpu'):
        """
        Forward pass of the normalization module that normalizes a given input tensor X.

        Parameters
        ----------
        X : torch.Tensor
            The input tensor to be normalized. Can be either a 3D or 4D tensor.
        device : str, optional
            The device to which the normalized tensor is transferred. Default: 'cpu'.

        Returns
        -------
        torch.Tensor
            The normalized tensor, adjusted to have zero mean and unit variance along the specified dimensions,
            and transferred to the specified device.
        """
        if len(X.shape) == 2:
            means = torch.mean(X, axis=0)
            X = X - means
            stds = torch.std(X, axis=0)
            X = X / stds

        if len(X.shape) == 3:
            means = torch.mean(X, axis=(0, 1))
            X = X - means.unsqueeze(0).unsqueeze(0)
            stds = torch.std(X, axis=(0, 1))
            X = X / stds.unsqueeze(0).unsqueeze(0)
        elif len(X.shape) == 4:
            means = torch.mean(X, dim=(0, 1, 2))
            X = X - means.unsqueeze(0).unsqueeze(0).unsqueeze(0)
            stds = torch.std(X, dim=(0, 1, 2))
            X = X / stds.unsqueeze(0).unsqueeze(0).unsqueeze(0)
        X[torch.where(torch.isnan(X))]=0
        self.means = means
        self.stds = stds

        return X.to(device)
    
    def denorm(self, X):
        X = X*self.stds
        X = X+self.means

        return X





class normalize_adj(nn.Module):
    """
    A PyTorch module for normalizing adjacency matrices to facilitate operations in graph neural networks. 
    The normalization adjusts adjacency matrices to account for node degrees, enhancing the propagation 
    of features through the network. This class can handle both batched and single adjacency matrices.

    Parameters
    ----------
    dim : int, optional
        The dimension over which to perform normalization (not utilized in current implementation). Default: 0.
    """
    def __init__(self, dim = 0):
        super().__init__()
        self.dim = dim
    
    def forward(self, Adj, device='cpu'):
        """
        Forward pass of the normalize_adj module that computes a degree-normalized adjacency matrix 
        from the input adjacency matrix 'Adj'. 

        Parameters
        ----------
        Adj : torch.Tensor or np.array
            The input adjacency matrix, which can be a 2D matrix for a single graph or a 3D tensor for batch processing.
        device : str, optional
            The device to which the normalized adjacency matrix is transferred. Default: 'cpu'.

        Returns
        -------
        torch.Tensor
            The degree-normalized adjacency matrix, transferred to the specified device.
        """
        if Adj is None:
            return None
        try:
            Adj = Adj.numpy()
        except:
            pass

        if len(Adj.shape) > 2:
            A_wave = Adj
            for i in range(Adj.shape[0]):
                A = Adj[i].reshape(Adj.shape[1], Adj.shape[1])
                A = A + np.diag(np.ones(A.shape[0], dtype=np.float32))
                D = np.array(np.sum(A, axis=1)).reshape((-1,))
                D[D <= 10e-5] = 10e-5    # Prevent infs
                diag = np.reciprocal(np.sqrt(D))
                A_wave[i] = np.multiply(np.multiply(diag.reshape((-1, 1)), A),
                                    diag.reshape((1, -1))).reshape(Adj[i].shape)
        else:
            A = Adj
            A = A + np.diag(np.ones(A.shape[0], dtype=np.float32))
            D = np.array(np.sum(A, axis=1)).reshape((-1,))
            D[D <= 10e-5] = 10e-5    # Prevent infs
            diag = np.reciprocal(np.sqrt(D))
            A_wave = np.multiply(np.multiply(diag.reshape((-1, 1)), A),
                                diag.reshape((1, -1)))
            
        return torch.FloatTensor(A_wave).to(device)




class convert_to_frequency(nn.Module):
    """
    A PyTorch module for transforming time-domain data into frequency-domain representations using either
    FFT (Fast Fourier Transform) or STFT (Short Time Fourier Transform). This module is configurable to handle
    different lengths of FFT and hop sizes for STFT, making it flexible for various signal processing tasks.

    Parameters
    ----------
    ftype : str, optional
        The type of frequency transformation to perform, either 'fft' for Fast Fourier Transform or 'stft' for
        Short Time Fourier Transform. Default: "fft".
    n_fft : int, optional
        The window size for the FFT or STFT. Default: 8.
    """

    def __init__(self, ftype="fft", n_fft=8):
        super().__init__()
        self.ftype = ftype
        self.n_fft = n_fft
        self.hop_length = n_fft // 2
        self.win_length = n_fft


    def forward(self, data, **kwarg):
        """
        Applies the configured frequency transformation (FFT or STFT) to the input data.

        Parameters
        ----------
        data : torch.Tensor
            The input data tensor, expected to be a time-domain signal. It can be a 3D tensor (batch, channels, time)
            for 'fft' or a 4D tensor (batch, nodes, time, features) for 'stft'.
        **kwargs : dict
            Additional keyword arguments, such as 'device' for specifying the computation device.

        Returns
        -------
        torch.Tensor
            The frequency-domain representation of the input data. The output format depends on the transformation
            type ('fft' returns real parts of the FFT, 'stft' returns the magnitude of the STFT as a 5D tensor).
        """
        if self.ftype == "fft":
            return torch.fft.fft(data, dim=2).real.float()
        
        elif self.ftype == "stft":
            num_graph, num_node, timestep, features = data.shape
            stft_results = []
            window = torch.hann_window(self.win_length, device=kwarg['device'])
            
            for g in range(num_graph):
                for n in range(num_node):
                    for f in range(features):
                        signal = data[g, n, :, f]
                        stft_result = torch.stft(signal, n_fft=self.n_fft, 
                                                 hop_length=self.hop_length, win_length=self.win_length,
                                                 window=window, return_complex=True).real.float()
                        stft_results.append(stft_result)
            num_frequencies = stft_results[0].shape[-2]
            num_timeframes = stft_results[0].shape[-1]

            stft_results = torch.stack(stft_results)
            stft_results = stft_results.view(num_graph, num_node, features, num_frequencies, num_timeframes)
            
            return stft_results

    

class add_time_embedding(nn.Module):
    """
    A PyTorch module that appends time-based embeddings to each feature vector in the dataset. 
    It can generate embeddings using sinusoidal functions, optionally using a Fourier transform approach, 
    enhancing the temporal aspects of data for tasks such as time series forecasting or sequence modeling.

    Parameters
    ----------
    embedding_dim : int, optional
        The dimensionality of the time embeddings. Default: 13.
    fourier : bool, optional
        Specifies whether to use a Fourier transform-based approach for time embedding. Default: False.
    """
    def __init__(self, embedding_dim=13, fourier=False):
        super().__init__()
        self.embedding_dim = embedding_dim
        self.fourier = fourier

    def forward(self, data, **kwarg):
        """
        Generates and appends time-based embeddings to the input data tensor.

        Parameters
        ----------
        data : torch.Tensor
            The input data tensor, which can vary in dimensions depending on the application (e.g., batch, nodes, time).
        **kwargs : dict
            Additional keyword arguments such as 'device' for specifying the computation device.

        Returns
        -------
        torch.Tensor
            The input data tensor augmented with time-based embeddings. The resulting tensor includes an additional
            dimension for the embeddings, concatenated to the last dimension of the input data tensor.
        """
        num_graphs = data.shape[0]
        num_nodes = data.shape[1]
        timesteps = data.shape[2]
        
        x = torch.arange(timesteps).float().unsqueeze(0).unsqueeze(1).to(kwarg['device'])

        a = torch.div(torch.arange(0.0, self.embedding_dim), 2, rounding_mode='floor').to(kwarg['device']) * 2 / self.embedding_dim
        b = torch.matmul(x.unsqueeze(-1), (2 * np.pi / 1440) * a.unsqueeze(0)) if not self.fourier else \
            torch.matmul(x.unsqueeze(-1), (1e-4 ** a).unsqueeze(0))
        c = torch.zeros_like(b)

        c[:, :, 0::2] = b[:, :, 0::2].sin()
        c[:, :, 1::2] = b[:, :, 1::2].cos()
        c = c.expand(num_graphs, num_nodes, -1, -1)

        if len(data.shape) == 3:
            data = data.unsqueeze(-1)
            data_with_time_embeddings = torch.cat((data, c), dim=-1) 
            #data_with_time_embeddings = torch.permute(data_with_time_embeddings, (0, 1, 3, 2))
        else:
            data_with_time_embeddings = torch.cat((data, c), dim=-1) 
        return data_with_time_embeddings
    
    
class learnable_time_embedding(nn.Module):
    """
    A PyTorch module that appends learnable time embeddings to the input data, facilitating the capture
    of temporal dependencies in models that process sequences or time-series data. The embedding is learned 
    during the training process, allowing it to adapt to specific temporal patterns observed in the dataset.

    Parameters
    ----------
    timesteps : int
        The total number of timesteps in the data sequence, which defines the number of unique time indices.
    embedding_dim : int
        The dimensionality of each time embedding vector.
    """
    def __init__(self, timesteps=13, embedding_dim=13):
        super(learnable_time_embedding, self).__init__()
        self.add_embedding = nn.Embedding(timesteps, embedding_dim)

    def forward(self, data, **kwarg):
        """
        Appends learnable time-based embeddings to the input data tensor. The embeddings are added to each timestep, 
        augmenting the feature dimensions of the data.

        Parameters
        ----------
        data : torch.Tensor
            The input data tensor, typically including dimensions for batch, nodes, and time.
        **kwargs : dict
            Additional keyword arguments such as 'device' for specifying the computation device.

        Returns
        -------
        torch.Tensor
            The input data tensor augmented with time-based embeddings, expanding the last dimension to include
            the embeddings.
        """
        
        num_graphs = data.shape[0]
        num_nodes = data.shape[1]
        timesteps = data.shape[2]
        
        
        time_indices = torch.arange(timesteps).expand(num_graphs, num_nodes, timesteps).to(kwarg['device'])
        #print(time_indices.shape)
        #print(self.embedding(time_indices).shape)
        data_with_time_embeddings = torch.cat((data, self.add_embedding(time_indices).squeeze(3)), dim=-1) 

        #print(data_with_time_embeddings.shape)  
        return data_with_time_embeddings


class seasonality_and_trend_decompose(nn.Module):
    """
    A PyTorch module designed to decompose time-series data into seasonality and trend components.
    It supports both dynamic and static decomposition methods. Dynamic decomposition leverages a Fourier
    transform approach for seasonality and convolutional filters for trend extraction. Static decomposition
    utilizes a moving average approach.

    Parameters
    ----------
    decompose_type : str, optional
        The type of decomposition to perform. "dynamic" for Fourier and convolutional methods, "static" for
        moving average based decomposition. Default: "dynamic".
    moving_avg : int, optional
        The window size for the moving average in static decomposition. Default: 25.
    kernel_size : list of int, optional
        List of kernel sizes for convolutional filters in dynamic trend decomposition. Default: [4, 8, 12].

    """
    def __init__(self, decompose_type = "dynamic", moving_avg=25, kernel_size= [4, 8, 12]):
        super().__init__()
        self.decompose_type = decompose_type
        if self.decompose_type == "dynamic":
            self.seasonality_model = FourierLayer(pred_len=0, k=3)
            self.trend_model = series_decomp_multi(kernel_size=kernel_size)
        if self.decompose_type == "static":
            self.decompsition = series_decomp(moving_avg)

    def forward(self, data, **kwarg):
        """
        Decomposes the input data tensor into seasonality and trend components. The method of decomposition
        (dynamic or static) impacts the models and techniques used.

        Parameters
        ----------
        data : torch.Tensor
            The input data tensor, typically including dimensions for batch, nodes, and time.
        **kwargs : dict
            Additional keyword arguments such as 'device' for specifying the computation device.

        Returns
        -------
        list of torch.Tensor
            A list containing the seasonality and trend components of the input data, each as a separate tensor.
        """
        num_graphs = data.shape[0]
        num_nodes = data.shape[1]
        timesteps = data.shape[2]
        time_middle_data = data.permute(0, 2, 1).contiguous()
        if self.decompose_type == "dynamic":
            _, trend = self.trend_model(time_middle_data)
            seasonality, _ = self.seasonality_model(time_middle_data)
        if self.decompose_type == "static":
            seasonality, trend = self.decompsition(time_middle_data)
        return [seasonality, trend]



class calculate_dtw_matrix(nn.Module):
    """
    A PyTorch module designed to compute the Dynamic Time Warping (DTW) distance matrix between all pairs
    of time-series in a dataset. DTW is a method that calculates an optimal match between two given sequences
    with certain restrictions. The matrix is computed once and saved for future use to avoid redundant computations.

    Parameters
    ----------
    dataset_name : str
        The name of the dataset, used to save and retrieve the computed DTW matrix.
    """
    def __init__(self, dataset_name):
        super().__init__()
        self.dataset = dataset_name
    def forward(self, data, **kwarg):
        """
        Computes the Dynamic Time Warping (DTW) matrix for the input data. If a precomputed matrix exists
        in the cache, it loads that matrix; otherwise, it computes a new matrix and saves it.

        Parameters
        ----------
        data : np.ndarray
            The input data array where each row represents a time step and each column a time-series node.
        **kwargs : dict
            Additional keyword arguments not utilized in this method.

        Returns
        -------
        np.ndarray
            The computed or loaded DTW distance matrix, where each element (i, j) represents the DTW distance
            between the i-th and j-th time-series.
        """
        # data format -> np.narray
        all_time = data.shape[0]
        num_nodes = data.shape[1]

        cache_path = './dtw_' + self.dataset + '.npy'
        if os.path.exists(cache_path):
            dtw_matrix = np.load(cache_path)
            print('Loaded DTW matrix from {}'.format(cache_path))
        else:
            data_mean = data.reshape(all_time,num_nodes,1)
            dtw_matrix = np.zeros((num_nodes, num_nodes))
            for i in tqdm(range(num_nodes)):
                for j in range(i, num_nodes):
                    dtw_distance, _ = fastdtw(data_mean[:, i, :], data_mean[:, j, :], radius=6)
                    dtw_matrix[i][j] = dtw_distance

            for i in range(num_nodes):
                for j in range(i):
                    dtw_matrix[i][j] = dtw_matrix[j][i]

            np.save(cache_path, dtw_matrix)
            print('Saved DTW matrix to {}'.format(cache_path))

        return dtw_matrix



# this code is adopted from https://github.com/chnsh/DCRNN_PyTorch/tree/pytorch_scratch/model/pytorch

import numpy as np
import torch
import torch.nn as nn
import scipy.sparse as sp
from scipy.sparse import linalg
import torch.nn.functional as F

from .base import BaseModel



'''
utils
'''
def calculate_normalized_laplacian(adj):
    """
    # L = D^-1/2 (D-A) D^-1/2 = I - D^-1/2 A D^-1/2
    # D = diag(A 1)
    :param adj:
    :return:
    """
    #adj = adj.to("cpu")
    adj = adj.cpu().clone().numpy()
    adj = sp.coo_matrix(adj)
    d = np.array(adj.sum(1))
    d_inv_sqrt = np.power(d, -0.5).flatten()
    d_inv_sqrt[np.isinf(d_inv_sqrt)] = 0.
    d_mat_inv_sqrt = sp.diags(d_inv_sqrt)
    normalized_laplacian = sp.eye(adj.shape[0]) - adj.dot(d_mat_inv_sqrt).transpose().dot(d_mat_inv_sqrt).tocoo()
    return normalized_laplacian

def calculate_scaled_laplacian(adj_mx, lambda_max=2, undirected=True):
    #print(adj_mx)
    #adj_mx = adj_mx.to("cpu")
    #print(adj_mx.shape)
    if undirected:
        adj_mx = torch.max(adj_mx, adj_mx.transpose(0, 1))
        #adj_mx = np.maximum.reduce([adj_mx, adj_mx.T])
        #adj_mx = np.maximum(adj_mx, adj_mx.T)
    L = calculate_normalized_laplacian(adj_mx)
    if lambda_max is None:
        lambda_max, _ = linalg.eigsh(L, 1, which='LM')
        lambda_max = lambda_max[0]
    L = sp.csr_matrix(L)
    M, _ = L.shape
    I = sp.identity(M, format='csr', dtype=L.dtype)
    L = (2 / lambda_max * L) - I
    return L.astype(np.float32)

def calculate_random_walk_matrix(adj_mx):
    #adj_mx = adj_mx.to("cpu")
    adj_mx = adj_mx.cpu().clone().numpy()
    adj_mx = sp.coo_matrix(adj_mx)
    d = np.array(adj_mx.sum(1))
    d_inv = np.power(d, -1).flatten()
    d_inv[np.isinf(d_inv)] = 0.
    d_mat_inv = sp.diags(d_inv)
    random_walk_mx = d_mat_inv.dot(adj_mx).tocoo()
    return random_walk_mx



'''
pytorch version of DCRNN written by Chintan Shah.
https://github.com/chnsh/DCRNN_PyTorch

This is a revised version of Chintan Shah's.
'''

'''
dcrnn_cell.py
'''
class LayerParams:
    def __init__(self, rnn_network: nn.Module, layer_type: str, device="cpu"):
        self._rnn_network = rnn_network
        self._params_dict = {}
        self._biases_dict = {}
        self._type = layer_type
        self.device = device

    def get_weights(self, shape):
        if shape not in self._params_dict:
            nn_param = nn.Parameter(torch.empty(*shape, device=self.device))
            nn.init.xavier_normal_(nn_param)
            self._params_dict[shape] = nn_param
            self._rnn_network.register_parameter('{}_weight_{}'.format(self._type, str(shape)),
                                                 nn_param)
        return self._params_dict[shape]

    def get_biases(self, length, bias_start=0.0):
        if length not in self._biases_dict:
            biases = torch.nn.Parameter(torch.empty(length, device=self.device))
            torch.nn.init.constant_(biases, bias_start)
            self._biases_dict[length] = biases
            self._rnn_network.register_parameter('{}_biases_{}'.format(self._type, str(length)),
                                                 biases)

        return self._biases_dict[length]


class DCGRUCell(nn.Module):
    def __init__(self, num_units, max_diffusion_step, nonlinearity='tanh',
                 filter_type="laplacian", device="cpu"):
        """
        :param num_units:
        :param adj_mx:
        :param max_diffusion_step:
        :param num_nodes:
        :param nonlinearity:
        :param filter_type: "laplacian", "random_walk", "dual_random_walk".
        """

        super().__init__()
        if nonlinearity == 'tanh':
            self._activation = torch.tanh 
        elif nonlinearity == 'relu':
            self._activation = torch.relu
        else:
            raise ValueError("Please specify an activation function in [tanh, relu]!") 
        # support other nonlinearities up here?
        # self._num_nodes = num_nodes
        self._num_units = num_units
        self._max_diffusion_step = max_diffusion_step
        self._supports = []
        self.device = device
        self.filter_type = filter_type

        #self._gconv_params = LayerParams(self, 'gconv', device=device)

    def reset_parameters(self):
        self._gconv_params = LayerParams(self, 'gconv', device=self.device)
            
            
    @staticmethod
    def _build_sparse_matrix(self, L):
        L = L.tocoo()
        indices = np.column_stack((L.row, L.col))
        # this is to ensure row-major ordering to equal torch.sparse.sparse_reorder(L)
        indices = indices[np.lexsort((indices[:, 0], indices[:, 1]))]
        L = torch.sparse_coo_tensor(indices.T, L.data, L.shape, device=self.device)
        return L

    def forward(self, inputs, hx, adj_mx, num_nodes):
        """Gated recurrent unit (GRU) with Graph Convolution.
        :param inputs: (B, num_nodes * num_features)
        :param hx: (B, num_nodes * rnn_units)

        :return
        - Output: A `2-D` tensor with shape `(B, num_nodes * rnn_units)`.
        """
        self._num_nodes = num_nodes
        if len(self._supports) == 0:
            supports = []
            if self.filter_type == "laplacian":
                supports.append(calculate_scaled_laplacian(adj_mx, lambda_max=None))
            elif self.filter_type == "random_walk":
                supports.append(calculate_random_walk_matrix(adj_mx).T)
            elif self.filter_type == "dual_random_walk":
                supports.append(calculate_random_walk_matrix(adj_mx).T)
                supports.append(calculate_random_walk_matrix(adj_mx.T).T)
            else:
                supports.append(calculate_scaled_laplacian(adj_mx))
            for support in supports:
                self._supports.append(self._build_sparse_matrix(self, support))
        
        output_size = 2 * self._num_units
        fn = self._gconv
        value = torch.sigmoid(fn(inputs, hx, output_size, bias_start=1.0))
        value = torch.reshape(value, (-1, num_nodes, output_size))
        r, u = torch.split(tensor=value, split_size_or_sections=self._num_units, dim=-1)
        r = torch.reshape(r, (-1, num_nodes * self._num_units))
        u = torch.reshape(u, (-1, num_nodes * self._num_units))

        c = self._gconv(inputs, r * hx, self._num_units)
        if self._activation is not None:
            c = self._activation(c)

        new_state = u * hx + (1.0 - u) * c
        return new_state

    @staticmethod
    def _concat(x, x_):
        x_ = x_.unsqueeze(0)
        return torch.cat([x, x_], dim=0)


    def _gconv(self, inputs, state, output_size, bias_start=0.0):
        # Reshape input and state to (batch_size, num_nodes, num_features/state_dim)
        batch_size = inputs.shape[0]
        inputs = torch.reshape(inputs, (batch_size, self._num_nodes, -1))
        state = torch.reshape(state, (batch_size, self._num_nodes, -1))
        inputs_and_state = torch.cat([inputs, state], dim=2)
        input_size = inputs_and_state.size(2)

        x = inputs_and_state
        x0 = x.permute(1, 2, 0)  # (num_nodes, total_arg_size, batch_size)
        x0 = torch.reshape(x0, shape=[self._num_nodes, input_size * batch_size])
        x = torch.unsqueeze(x0, 0)

        if self._max_diffusion_step == 0:
            pass
        else:
            for support in self._supports:
                x1 = torch.sparse.mm(support, x0)
                x = self._concat(x, x1)

                for k in range(2, self._max_diffusion_step + 1):
                    x2 = 2 * torch.sparse.mm(support, x1) - x0
                    x = self._concat(x, x2)
                    x1, x0 = x2, x1

        num_matrices = len(self._supports) * self._max_diffusion_step + 1  # Adds for x itself.
        x = torch.reshape(x, shape=[num_matrices, self._num_nodes, input_size, batch_size])
        x = x.permute(3, 1, 2, 0)  # (batch_size, num_nodes, input_size, order)
        x = torch.reshape(x, shape=[batch_size * self._num_nodes, input_size * num_matrices])

        weights = self._gconv_params.get_weights((input_size * num_matrices, output_size))
        x = torch.matmul(x, weights)  # (batch_size * self._num_nodes, output_size)

        biases = self._gconv_params.get_biases(output_size, bias_start)
        x += biases
        # Reshape res back to 2D: (batch_size, num_node, state_dim) -> (batch_size, num_node * state_dim)
        return torch.reshape(x, [batch_size, self._num_nodes * output_size])
    
    
    
    

'''
dcrnn_model.py
'''
def count_parameters(model):
    return sum(p.numel() for p in model.parameters() if p.requires_grad)


class Seq2SeqAttrs:
    def __init__(self, max_diffusion_step, filter_type,
                              num_rnn_layers, rnn_units):
        self.max_diffusion_step = max_diffusion_step
        self.filter_type = filter_type
        self.num_rnn_layers = num_rnn_layers
        self.rnn_units = rnn_units


class EncoderModel(nn.Module, Seq2SeqAttrs):
    def __init__(self, 
                 max_diffusion_step, filter_type, num_rnn_layers, rnn_units,
                 num_features, num_timesteps_input, nonlinearity='tanh', device="cpu"):
        nn.Module.__init__(self)
        Seq2SeqAttrs.__init__(self, max_diffusion_step, filter_type,
                              num_rnn_layers, rnn_units)
        self.num_features = num_features
        self.num_timesteps_input = num_timesteps_input
        self.dcgru_layers = nn.ModuleList(
            [DCGRUCell(self.rnn_units, self.max_diffusion_step, nonlinearity=nonlinearity,
                       filter_type=self.filter_type, device=device) for _ in range(self.num_rnn_layers)])
        self.device = device

    def forward(self, inputs, hidden_state=None, adj_mx=None, num_nodes=1, dropout=0):
        """
        Encoder forward pass.

        :param inputs: shape (batch_size, self.num_nodes * self.num_features)
        :param hidden_state: (num_layers, batch_size, self.hidden_state_size)
               optional, zeros if not provided
        :return: output: # shape (batch_size, self.hidden_state_size)
                 hidden_state # shape (num_layers, batch_size, self.hidden_state_size)
                 (lower indices mean lower layers)
        """
        hidden_state_size = num_nodes * self.rnn_units
        batch_size, _ = inputs.size()
        if hidden_state is None:
            hidden_state = torch.zeros((self.num_rnn_layers, batch_size, hidden_state_size),
                                       device=self.device)
        hidden_states = []
        output = inputs
        for layer_num, dcgru_layer in enumerate(self.dcgru_layers):
            next_hidden_state = dcgru_layer(output, hidden_state[layer_num], adj_mx, num_nodes)
            hidden_states.append(next_hidden_state)
            output = next_hidden_state
            if layer_num != len(self.dcgru_layers) - 1:
                output = F.dropout(output, p=dropout, training=self.training)

        return output, torch.stack(hidden_states)  # runs in O(num_layers) so not too slow
    
    def initialize(self):
        for m in self.dcgru_layers:
            m.reset_parameters()


class DecoderModel(nn.Module, Seq2SeqAttrs):
    def __init__(self,
                 max_diffusion_step, filter_type, num_rnn_layers, rnn_units,
                 num_classes, num_timesteps_output, nonlinearity='tanh', device="cpu"):
        # super().__init__(is_training, adj_mx, **model_kwargs)
        nn.Module.__init__(self)
        Seq2SeqAttrs.__init__(self, max_diffusion_step, filter_type,
                              num_rnn_layers, rnn_units)
        self.num_classes = num_classes
        self.num_timesteps_output = num_timesteps_output
        self.projection_layer = nn.Linear(self.rnn_units, self.num_classes)
        self.dcgru_layers = nn.ModuleList(
            [DCGRUCell(self.rnn_units, self.max_diffusion_step,
                       filter_type=self.filter_type, nonlinearity=nonlinearity, device=device) for _ in range(self.num_rnn_layers)])

    def forward(self, inputs, hidden_state=None, adj_mx=None, num_nodes=1, dropout=0):
        """
        Decoder forward pass.

        :param inputs: shape (batch_size, self.num_nodes * self.num_classes)
        :param hidden_state: (num_layers, batch_size, self.hidden_state_size)
               optional, zeros if not provided
        :return: output: # shape (batch_size, self.num_nodes * self.num_classes)
                 hidden_state # shape (num_layers, batch_size, self.hidden_state_size)
                 (lower indices mean lower layers)
        """
        hidden_states = []
        output = inputs
        for layer_num, dcgru_layer in enumerate(self.dcgru_layers):
            next_hidden_state = dcgru_layer(output, hidden_state[layer_num], adj_mx, num_nodes)
            hidden_states.append(next_hidden_state)
            output = next_hidden_state
            if layer_num != len(self.dcgru_layers) - 1:
                output = F.dropout(output, p=dropout, training=self.training)

        projected = self.projection_layer(output.view(-1, self.rnn_units))
        output = projected.view(-1, num_nodes * self.num_classes)

        return output, torch.stack(hidden_states)
    
    
    def initialize(self):
        for m in self.dcgru_layers:
            m.reset_parameters()


class DCRNN(BaseModel, Seq2SeqAttrs):
    """
    Diffusion Convolutional Recurrent Neural Network (DCRNN) 

    Parameters
    ----------
    num_features : int
        Number of input features per node.
    num_timesteps_input : int
        Number of past time steps used as input by the network.
    num_classes : int
        Number of output features per node.
    num_timesteps_output : int
        Number of future time steps to predict.
    max_diffusion_step : int
        Maximum number of diffusion steps in the graph convolution operations. Default: 2.
    filter_type : str
        Type of filter used in graph convolutions, e.g., 'laplacian'. Default: "laplacian".
    num_rnn_layers : int
        Number of recurrent neural network layers. Default: 1.
    rnn_units : int
        Number of units per recurrent layer. Default: 1.
    nonlinearity : str
        Type of nonlinearity function used in RNN. Default: "tanh".
    dropout : float
        Dropout rate applied in the network to prevent overfitting. Default: 0.
    device : str, optional
        The device (cpu or gpu) on which the model will be run. Default: 'cpu'.

    Returns
    -------
    torch.Tensor
        A tensor of shape (batch_size, num_nodes, horizon), representing the predicted values for each node over future timesteps.
    """
    def __init__(self,
              num_features=1,
              num_timesteps_input=5,
              num_classes=1,
              num_timesteps_output=1,
              max_diffusion_step=2, 
              filter_type="laplacian",
              num_rnn_layers=1, 
              rnn_units=1,
              nonlinearity="tanh",
              dropout=0,
              device="cpu"):

        super().__init__()
        Seq2SeqAttrs.__init__(self, max_diffusion_step=max_diffusion_step, filter_type=filter_type,
                              num_rnn_layers=num_rnn_layers, rnn_units=rnn_units)
        
        
        self.encoder_model = EncoderModel(max_diffusion_step=max_diffusion_step, filter_type=filter_type,
                                          num_rnn_layers=num_rnn_layers, rnn_units=rnn_units,
                                          num_features=num_features, num_timesteps_input=num_timesteps_input, nonlinearity=nonlinearity, device=device)
        self.decoder_model = DecoderModel(max_diffusion_step=max_diffusion_step, filter_type=filter_type,
                                          num_rnn_layers=num_rnn_layers, rnn_units=rnn_units,
                                          num_classes=num_classes, num_timesteps_output=num_timesteps_output, nonlinearity=nonlinearity, device=device)
        
        self.device = device
        self.num_features = num_features
        self.num_timesteps_input = num_timesteps_input
        self.num_classes = num_classes
        self.num_timesteps_output = num_timesteps_output
        self.dropout = dropout



    def encoder(self, inputs, adj_mx, num_nodes):
        """
        encoder forward pass on t time steps
        :param inputs: shape (num_timesteps_input, batch_size, num_sensor * num_features)
        :return: encoder_hidden_state: (num_layers, batch_size, self.hidden_state_size)
        """
        encoder_hidden_state = None
        for t in range(self.encoder_model.num_timesteps_input):
            _, encoder_hidden_state = self.encoder_model(inputs[t], encoder_hidden_state, adj_mx, num_nodes, self.dropout)

        return encoder_hidden_state

    def decoder(self, encoder_hidden_state, adj_mx, num_nodes):
        """
        Decoder forward pass
        :param encoder_hidden_state: (num_layers, batch_size, self.hidden_state_size)
        :param labels: (self.num_timesteps_output, batch_size, self.num_nodes * self.num_classes) [optional, not exist for inference]
        :param batches_seen: global step [optional, not exist for inference]
        :return: output: (self.num_timesteps_output, batch_size, self.num_nodes * self.num_classes)
        """
        batch_size = encoder_hidden_state.size(1)
        go_symbol = torch.zeros((batch_size, num_nodes * self.decoder_model.num_classes),
                                device=self.device)
        decoder_hidden_state = encoder_hidden_state
        decoder_input = go_symbol

        outputs = []

        for t in range(self.decoder_model.num_timesteps_output):
            decoder_output, decoder_hidden_state = self.decoder_model(decoder_input,
                                                                      decoder_hidden_state, adj_mx, num_nodes, self.dropout)
            decoder_input = decoder_output
            outputs.append(decoder_output)

        outputs = torch.stack(outputs)
        return outputs

    def forward(self, X_batch, graph, X_states, batch_graph):
        """
        Parameters
        ----------
        X_batch : torch.Tensor
            Input tensor with shape (batch_size, num_nodes, num_timesteps_input, num_features),
            representing the input features over multiple timesteps for each node.
        graph : torch.Tensor
            Static adjacency matrix with shape (num_nodes, num_nodes),
            representing the fixed connections between nodes.
        X_states : torch.Tensor, optional
            States of the nodes if available, with the same shape as X_batch.
            Used for models that incorporate node states over time. Default: None.
        batch_graph : torch.Tensor, optional
            Dynamic adjacency matrix if available, with shape similar to graph but possibly varying over time.
            Used for models that account for changing graph structures. Default: None.

        Returns
        -------
        torch.Tensor
            The output tensor of shape (batch_size, num_nodes, num_timesteps_output),
            representing the predicted values for each node over the specified output timesteps.
        """
        
        inputs = X_batch
        batch_size = X_batch.shape[0] #data.batch[-1] + 1
        
        # reshape geometric dataloader format to real format
        '''num_graphs = data.num_graphs
        individual_graphs = []
        for i in range(num_graphs):
            mask = data.batch == i
            node_features = data.x[mask]
            individual_graphs.append(node_features)
            adj_m = data.adj_m[mask]
        new_inputs = torch.stack(individual_graphs, dim=0)'''
        
        inputs = torch.permute(inputs,(2, 0, 1, 3))
        inputs = torch.reshape(inputs, (inputs.shape[0], inputs.shape[1], 
                                            inputs.shape[2]*inputs.shape[3]))
        
        
        num_nodes = graph.shape[0]
        encoder_hidden_state = self.encoder(inputs, graph, num_nodes)
        outputs = self.decoder(encoder_hidden_state, graph, num_nodes)
        
        outputs = torch.reshape(outputs, (self.num_timesteps_output, batch_size, 
                                            num_nodes, self.num_classes))
        outputs = torch.permute(outputs, (1, 2, 0, 3)).squeeze(-1)
        #outputs = torch.reshape(outputs, (-1, outputs.shape[2], outputs.shape[3]))
        return outputs
    
    
    def initialize(self):
        self.encoder_model.initialize()
        self.decoder_model.initialize()
        #pass


    



# this code is adopted from https://github.com/DSE-MSU/DeepRobust/blob/master/deeprobust/graph/defense_pyg/gcn.py

import torch.nn as nn
import torch.nn.functional as F
from torch_geometric.nn import GCNConv
import torch

from .base import BaseModel

class GCN(BaseModel):
    """
    Graph Convolutional Network (GCN)

    Parameters
    ----------
    num_features : int
        Number of input features per node.
    hidden_dim : int, optional
        Dimension of hidden layers. Default: 16.
    num_classes : int, optional
        Number of output classes for each node. Default: 2.
    nlayers : int, optional
        Number of layers in the GCN. Default: 2.
    dropout : float, optional
        Dropout rate for regularization during training to prevent overfitting. Default: 0.5.
    with_bn : bool, optional
        Specifies whether batch normalization should be included. Default: False.
    with_bias : bool, optional
        Specifies whether to include bias parameters in the GCN layers. Default: True.
    device : str
        The device (cpu or gpu) on which the model will be run.

    Returns
    -------
    torch.Tensor
        A tensor of shape (batch_size, num_nodes, num_classes), representing the predicted outcomes for each node after passing through the GCN.
    """
    def __init__(self, num_features, hidden_dim=16, num_classes=2, nlayers=2, dropout=0.5,
                with_bn=False, with_bias=True, device='cpu'):

        super(GCN, self).__init__()

        assert device is not None, "Please specify 'device'!"
        self.device = device
        self.hid = hidden_dim
        self.out_dim = num_classes

        self.layers = nn.ModuleList([])
        if with_bn:
            self.bns = nn.ModuleList()


        if nlayers == 1:
            self.layers.append(GCNConv(num_features, hidden_dim, bias=with_bias))
        else:
            self.layers.append(GCNConv(num_features, hidden_dim, bias=with_bias))
            if with_bn:
                self.bns.append(nn.BatchNorm1d(hidden_dim))
            for i in range(nlayers-2):
                self.layers.append(GCNConv(hidden_dim, hidden_dim, bias=with_bias))
                if with_bn:
                    self.bns.append(nn.BatchNorm1d(hidden_dim))
            self.layers.append(GCNConv(hidden_dim, hidden_dim, bias=with_bias))
        
        self.fc = nn.Linear(hidden_dim, num_classes)
        self.dropout = dropout
        self.with_bn = with_bn

    def forward(self, x, edge_index, edge_weight):
        """
        Parameters:
        x : torch.Tensor
            The input features tensor with shape (batch_size, num_nodes, num_features).
        edge_index : torch.Tensor
            The edge indices in COO format with shape (2, num_edges).
            
        Returns:
        torch.Tensor
            The output predictions for each node with shape (batch_size * num_nodes, num_classes).
        """
        #print(x.shape)
        #b, n, _= x.shape
        x = torch.reshape(x, (x.shape[0], -1))
        #print(x.shape)
    
        for i, layer in enumerate(self.layers):
            if edge_weight is not None:
                x = layer(x.float(), edge_index=edge_index, edge_weight=edge_weight)
            else:
                x = layer(x.float(), edge_index)
            if i != len(self.layers) - 1:
                if self.with_bn:
                    x = self.bns[i](x)#(x.view(-1, self.hid)).view(b, n, self.hid)
                x = F.relu(x)
                x = F.dropout(x, p=self.dropout, training=self.training)
        return self.fc(x) #self.fc(x)#.squeeze() #F.log_softmax(x, dim=1)

    def initialize(self):
        for m in self.layers:
            m.reset_parameters()
        if self.with_bn:
            for bn in self.bns:
                bn.reset_parameters()



import torch.nn as nn
import torch.nn.functional as F
from torch_geometric.nn import GINConv
from torch_geometric.nn.models import MLP
import torch

from .base import BaseModel

class GIN(BaseModel):
    """
    Graph Isomorphism Network (GIN)

    Parameters
    ----------
    num_features : int
        Number of input features per node.
    hidden_dim : int
        Dimension of hidden layers.
    num_classes : int
        Number of output features per node.
    nlayers : int, optional
        Number of layers in the GIN. Default: 2.
    dropout : float, optional
        Dropout rate for regularization during training to prevent overfitting. Default: 0.5.
    with_bias : bool, optional
        Specifies whether to include bias parameters in the MLP layers. Default: True.
    device : str
        The device (cpu or gpu) on which the model will be run. Must be specified.

    Returns
    -------
    torch.Tensor
        A tensor of shape (batch_size, num_nodes, output_dim), representing the predicted outcomes for each node after passing through the GIN. 
    """
    def __init__(self, num_features, hidden_dim, num_classes, nlayers=2, dropout=0.5,
                with_bias=True, device=None):

        super(GIN, self).__init__()

        assert device is not None, "Please specify 'device'!"
        self.device = device

        
        mlp_one = MLP(
            [num_features, hidden_dim, num_classes],
            act="relu",
            norm="batch_norm",
            bias=with_bias
        )
        
        mlp_first = MLP(
            [num_features, hidden_dim, hidden_dim],
            act="relu",
            norm="batch_norm",
            bias=with_bias
        )
        
        mlp_hidden = MLP(
            [hidden_dim, hidden_dim, hidden_dim],
            act="relu",
            norm="batch_norm",
            bias=with_bias
        )
        
        mlp_last = MLP(
            [hidden_dim, hidden_dim, num_classes],
            act="relu",
            norm="batch_norm",
            bias=with_bias
        )
        
        self.layers = nn.ModuleList([])

        if nlayers == 1:
            self.layers.append(GINConv(mlp_first))
        else:
            self.layers.append(GINConv(mlp_first))
            for i in range(nlayers-2):
                self.layers.append(GINConv(mlp_hidden))
            self.layers.append(GINConv(mlp_hidden))
        
        self.fc = nn.Linear(hidden_dim, num_classes)
        self.dropout = dropout


    def forward(self, x, edge_index, edge_weight):
        """
        Parameters:
        x : torch.Tensor
            The input features tensor with shape (batch_size, num_nodes, num_features).
        edge_index : torch.Tensor
            The edge indices in COO format with shape (2, num_edges).
            
        Returns:
        torch.Tensor
            The output predictions for each node with shape (batch_size * num_nodes, num_classes).
        """
        #x, edge_index = data.x, data.edge_index
        x = torch.reshape(x, (x.shape[0], -1))
        for i, layer in enumerate(self.layers):
            x = layer(x, edge_index)
            if i != len(self.layers) - 1:
                x = F.relu(x)
                x = F.dropout(x, p=self.dropout, training=self.training)
        return self.fc(x) #F.log_softmax(x, dim=1)

    def initialize(self):
        for m in self.layers:
            m.reset_parameters()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="epilearn",
    version="0.0.6",
    author="Emory-Melody", 
    maintainer='Emory-Melody',
    author_email="zevin.liu@gmail.com",
    description="A Pytorch library for machine learning in epidemic modeling", 
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Emory-Melody/EpiLearn",
    include_package_data=True,
    install_requires = [
          'matplotlib==3.9.1',
          'numpy==1.26.4',
          'scipy>=1.3.1',
          'networkx==3.2.1',
          'Pillow>=7.0.0',
          'scikit_learn>=0.22.1',
          'tqdm>=3.0',
          'seaborn>=0.13.2',
          'xgboost>=2.0.3',
          'statsmodels>=0.14.2',
          'streamlit==1.34.0',
          'pyvis==0.3.2',
          'plotly==5.22.0',
          'fastdtw==0.3.4',
          'einops==0.8.0' 
      ],
    packages=setuptools.find_packages(),
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Libraries",
    ],
    license="MIT",
)


# this code is adopted from https://github.com/littlecherry-art/DASTGN/blob/master/Code/model.py

import torch
import torch.nn as nn
import torch.nn.functional as F
import math

from ...utils.utils import Degree_Matrix, Static_full, kronecker
from .base import BaseModel


#--------------For Fine Module---------------

class TimeEffect(nn.Module):
    def __init__(self, node_hid_size, out_size, num_timestamps, num_nodes):
        """
        Time-specific effect: Compute the fixed spatial dependency of each specific time slice
        """
        super(TimeEffect, self).__init__()

        self.gru = nn.GRU(input_size=node_hid_size, hidden_size=out_size, batch_first=True)

        self.num_timestamps = num_timestamps
        self.tot_nodes = num_nodes
        self.dim = self.num_timestamps * self.tot_nodes
        self.input_size = node_hid_size

    def forward(self, raw_features):
        node_features = raw_features.transpose(0, 1)
        temp_out, temp_hid = self.gru(node_features)

        temp = temp_out.transpose(0, 1)
        spatial_scores = torch.matmul(temp, temp.transpose(1, 2).contiguous()) / math.sqrt(self.input_size)
        spatial_weights = torch.sigmoid(spatial_scores)

        ST_weights = spatial_weights.repeat(self.num_timestamps, 1, 1)
        ST_weights = ST_weights.view(self.num_timestamps, self.num_timestamps, self.tot_nodes, self.tot_nodes)
        ST_weights = ST_weights.permute(0, 2, 1, 3)              # affected by fixed spatial dependency of each historical time slice
        ST_weights = ST_weights.reshape(self.dim, self.dim)

        return ST_weights


'''
Space-specific effect: Compute the fixed temporal weights at each specific region
'''
class GAT(nn.Module):

    def __init__(self, input_size):
        super(GAT, self).__init__()

        self.input_size = input_size

    def forward(self, input, adj):

        spatial_scores = torch.matmul(input, input.transpose(1, 2))/math.sqrt(self.input_size)
        zero_vec = -1e12 * torch.ones_like(spatial_scores)
        attention = torch.where(adj > 0, spatial_scores, zero_vec)
        attention = F.softmax(attention, dim=2)
        out_hid = torch.matmul(attention, input)

        return out_hid


class SpaceEffect(nn.Module):

    def __init__(self, node_input_size, num_nodes, num_timestamps):
        super(SpaceEffect, self).__init__()

        self.gat = GAT(node_input_size)

        self.num_timestamps = num_timestamps
        self.tot_nodes = num_nodes
        self.dim = self.num_timestamps * self.tot_nodes
        self.input_size = node_input_size

    def forward(self, raw_features, adj):
        node_features = self.gat(raw_features, adj)

        ## temporal dependency at each specific region
        spatial_features = node_features.transpose(0, 1)
        temp_scores = torch.matmul(spatial_features, spatial_features.transpose(1, 2).contiguous())/math.sqrt(self.input_size)
        spatial_weights = torch.sigmoid(temp_scores)

        ST_weights = spatial_weights.repeat(self.tot_nodes, 1, 1)
        ST_weights = ST_weights.view(self.tot_nodes, self.tot_nodes, self.num_timestamps, self.num_timestamps)
        ST_weights = ST_weights.permute(3, 1, 2, 0)            # affected by temporal dependency of self region
        ST_weights = ST_weights.reshape(self.dim, self.dim)

        return ST_weights


'''
Direct interaction effect: directly compute the weight between any two space-time points
'''
class DirectEffect(nn.Module):

    def __init__(self, input_size, num_nodes, num_timestamps):
        super(DirectEffect, self).__init__()

        self.num_timestamps = num_timestamps
        self.tot_nodes = num_nodes
        self.input_size = input_size


    def forward(self, raw_features):
        node_features = raw_features.view(raw_features.size()[0] * raw_features.size()[1], -1)
        st_scores = torch.matmul(node_features, node_features.transpose(0, 1).contiguous()) / math.sqrt(self.input_size)
        st_scores = torch.sigmoid(st_scores)

        return st_scores


'''
Integrate three NT*NT dimensional spatio-temporal weight matrices generated by different space-time effects
'''
class MultiEffectFusion(nn.Module):

    def __init__(self, input_size, num_nodes, num_timestamps):
        super(MultiEffectFusion, self).__init__()

        self.input_size = input_size
        self.tot_nodes = num_nodes
        self.num_timestamps = num_timestamps
        self.dim = num_timestamps * self.tot_nodes

        self.Time = TimeEffect(input_size, input_size, num_timestamps, num_nodes)
        self.Space = SpaceEffect(input_size, num_nodes, num_timestamps)
        self.Direct = DirectEffect(input_size, num_nodes, num_timestamps)

        self.TW = nn.Parameter(torch.FloatTensor(1, 1))
        self.SW = nn.Parameter(torch.FloatTensor(1, 1))
        self.DW = nn.Parameter(torch.FloatTensor(1, 1))


    def forward(self, his_raw_features, adj):     # dim(his_raw_features)=(timestamp, num_vertices, num_features)
        filter = Static_full(self.tot_nodes, self.num_timestamps, adj)

        temporal_matrix = self.Time(his_raw_features) * filter
        temporal_gate = self.TW * temporal_matrix

        spatial_matrix = self.Space(his_raw_features, adj) * filter
        spatial_gate = self.SW * spatial_matrix

        direct_matrix = self.Direct(his_raw_features) * filter
        direct_gate = self.DW * direct_matrix

        gates = torch.cat((temporal_gate.unsqueeze(0), spatial_gate.unsqueeze(0), direct_gate.unsqueeze(0)), 0)
        gates = F.softmax(gates, dim=0)

        ST_matrix = gates[0, :, :] * temporal_matrix + gates[1, :, :] * spatial_matrix + gates[2, :, :] * direct_matrix

        return ST_matrix


#--------------For Coarase Module---------------


'''
A general function to construct space-time neighboring blocks / space-time dependency structures
'''
def Static(n, t, A, rho_IT, rho_CT1, rho_CT2):
    """
    :param n: the dimension of the spatial adjacency matrix
    :param t: the length of periods
    :param A: the spatial adjacency matrix
    :param rho_IT: the trainable paramter of the current states of neighbors
    :param rho_CT1: the trainable paramter of the historical states of self
    :param rho_CT2: the trainable paramter of the historical states of neighbors
    :return: a space-time dependency structure matrix
    """
    I_S = torch.diag_embed(torch.ones(n))
    I_T = torch.diag_embed(torch.ones(t))

    C_S = A
    C_T = torch.tril(torch.ones(t, t), diagonal=-1)

    A_ST = kronecker(rho_IT * I_T, C_S) + kronecker(rho_CT1 * C_T, I_S) + kronecker(rho_CT2 * C_T, C_S)
    # A_ST = rho_CT1 * kronecker(C_T, I_S) + rho_CT2 * kronecker(C_T, C_S) + rho_IT * kronecker(I_T, C_S)

    return A_ST


'''
Coarse-grained module based on interventions
'''
class STNB_layer(nn.Module):
    # return NT * NT dimensional weight matrix for each type of space-time neighbor blocks
    def __init__(self, tot_nodes, num_timestamps, input_size):
        super(STNB_layer, self).__init__()
        self.dim = tot_nodes * num_timestamps
        self.num_timestamps = num_timestamps

        self.w1 = nn.Parameter(torch.FloatTensor(1, 1))
        self.w2 = nn.Parameter(torch.FloatTensor(1, 1))
        self.gate = nn.Sequential(nn.Linear(input_size, 1), nn.Sigmoid())

    def forward(self, features, interven, block_matrix, adj):
        judge = torch.sum(adj) * self.num_timestamps
        # process cumulative/current intervention
        block_sum = torch.sum(block_matrix)
        if block_sum > judge:
            interven_cum = torch.cumsum(interven, dim=0) - interven           # cumulative historical interventions
            interven_adjust = interven_cum.squeeze().view(self.dim, 1)
        else:
            interven_adjust = interven.squeeze().view(self.dim, 1)            # current intervention


        Infor = torch.mm(block_matrix, features)
        feat = self.w1 * Infor + interven_adjust.repeat(1,features.shape[1]) + self.w2 * features

        rho = self.gate(feat)
        # weights of all impact points in the same block are identical
        block_weight = torch.mul(rho.repeat(1, self.dim), block_matrix)

        return block_weight


class Coarse_module(nn.Module):
    # return NT * NT dimensional weight matrix of three types of space-time neighbor blocks
    def __init__(self, tot_nodes, num_timestamps, input_size):

        super(Coarse_module, self).__init__()
        self.tot_nodes = tot_nodes
        self.num_timestamps = num_timestamps
        self.input_size = input_size

        self.Gate_IT = STNB_layer(self.tot_nodes, num_timestamps, input_size)
        self.Gate_CS = STNB_layer(self.tot_nodes, num_timestamps, input_size)
        self.Gate_CT = STNB_layer(self.tot_nodes, num_timestamps, input_size)

    def forward(self, his_raw_features, interven, adj):
        # ipdb.set_trace()

        features = his_raw_features.contiguous().view(-1, self.input_size)
        A_IT = Static(self.tot_nodes, self.num_timestamps, adj, rho_IT=1, rho_CT1=0, rho_CT2=0)     # current states of neighbors
        A_CS = Static(self.tot_nodes, self.num_timestamps, adj, rho_IT=0, rho_CT1=1, rho_CT2=0)     # historical states of self
        A_CT = Static(self.tot_nodes, self.num_timestamps, adj, rho_IT=0, rho_CT1=0, rho_CT2=1)     # historical states of neighbors

        gate_IT = self.Gate_IT(features, interven, A_IT, adj)
        gate_CS = self.Gate_CS(features, interven, A_CS, adj)
        gate_CT = self.Gate_CT(features, interven, A_CT, adj)

        gate = gate_IT + gate_CS + gate_CT

        return gate

#--------------Downstream Model---------------

class Regression(nn.Module):

    def __init__(self, emb_size, out_size):
        super(Regression, self).__init__()

        self.layer = nn.Sequential(nn.Linear(emb_size, emb_size),
                                   nn.ReLU(),
                                   nn.Linear(emb_size, out_size),
                                   nn.ReLU())

        self.init_params()

    def init_params(self):
        for param in self.parameters():
            if len(param.size()) == 2:
                nn.init.xavier_uniform_(param)

    def forward(self, embds):
        logists = self.layer(embds)
        return logists


#--------------Main Model---------------

'''
Spatio-Temporal GNN
'''
class GConv(nn.Module):

    def __init__(self, D_temporal, A_temporal, tot_nodes, tw, fw):
        super(GConv, self).__init__()

        self.tot_nodes = tot_nodes
        self.sp_temp = torch.mm(D_temporal, torch.mm(A_temporal, D_temporal))

        self.his_temporal_weight = tw
        self.his_final_weight = fw

    def forward(self, his_raw_features):
        his_self = his_raw_features
        his_temporal = self.his_temporal_weight.repeat(self.tot_nodes, 1) * his_raw_features
        his_temporal = torch.mm(self.sp_temp, his_temporal)

        his_combined = torch.cat([his_self, his_temporal], dim=1)
        his_raw_features = torch.relu(his_combined.mm(self.his_final_weight))

        return his_raw_features
    

class DASTGN(BaseModel):
    """
    Dynamic and Adaptive Spatio-Temporal Graph Network (DASTGN) 
    
    Parameters
    ----------
    num_nodes : int
        Number of nodes in the graph.
    num_features : int
        Number of features per node per timestep.
    num_timesteps_input : int
        Number of timesteps considered for each input sample.
    num_timesteps_output : int
        Number of output timesteps to predict.
    GNN_layers : int, optional
        Number of Graph Neural Network layers to use. Default: 2.
    device : str, optional
        The device (cpu or gpu) on which the model will be run. Default: 'cpu'.

    Returns
    -------
    torch.Tensor
        A tensor of shape (batch_size, num_timesteps_output, num_nodes), representing the predicted values for each node over future timesteps.
        Each slice along the second dimension corresponds to a timestep, with each column representing a node.
    """
    def __init__(self, 
                num_nodes,
                num_features,
                num_timesteps_input,
                num_timesteps_output,
                GNN_layers = 2,
                device = 'cpu'):
        super(DASTGN, self).__init__()

        self.device = device
        self.num_timestamps = num_timesteps_input
        self.input_size = num_features
        self.tot_nodes = num_nodes
        self.device = device
        self.GNN_layers = GNN_layers
        self.dim = self.num_timestamps * self.tot_nodes
        self.embd_size = self.num_timestamps * self.input_size
        self.out_size = num_timesteps_output

        self.his_temporal_weight = nn.Parameter(torch.FloatTensor(self.num_timestamps, self.input_size))
        self.his_final_weight = nn.Parameter(torch.FloatTensor(2 * self.input_size, self.input_size))
        self.final_weight = nn.Parameter(torch.FloatTensor(self.embd_size, self.embd_size))

        # a trainable parameter of the time-decaying interventions after quarantine
        self.theta = nn.Parameter(torch.FloatTensor(1, 1))

        self.Coarse_module = Coarse_module(self.tot_nodes, self.num_timestamps, self.input_size)
        self.Fine_module = MultiEffectFusion(self.input_size, self.tot_nodes, self.num_timestamps)

        self.output_module = Regression(emb_size=self.embd_size, out_size=self.out_size) 

        self.init_params()

    def init_params(self):
        for param in self.parameters():
            if len(param.size()) == 2:
                nn.init.xavier_uniform_(param)

    def forward(self, x, adj, states=None, dynamic_adj=None, **kwargs):     # dim(his_raw_features)=(Time, Space, Feat)
        """
        Parameters
        ----------
        x : torch.Tensor
            Input features tensor with shape (batch_size, num_timesteps_input, num_nodes, num_features).
        adj : torch.Tensor
            Static adjacency matrix of the graph with shape (num_nodes, num_nodes).
        states : torch.Tensor, optional
            States of the nodes if available, with the same shape as x. Default: None.
        dynamic_adj : torch.Tensor, optional
            Dynamic adjacency matrix if available, with shape similar to adj but possibly varying over time. Default: None.

        Returns
        -------
        torch.Tensor
            The output tensor of shape (batch_size, num_timesteps_output, num_nodes),
            representing the predicted values for each node over the specified output timesteps.
        """
        batch, w, n, f = x.shape
        x = x[..., :self.input_size]              # delete time and intervention variables

        if self.input_size == x.shape[-1]:
            interven_adjust = torch.zeros_like(x)[..., :1]
        else:
            time_interval = x[..., self.input_size:(self.input_size+1)]    # construct a piece-wise function for intervention
            interven = x[..., (self.input_size+1):(self.input_size+2)]
            interven_decay = torch.sigmoid(-self.theta * time_interval)
            interven_adjust = torch.where(interven == 0.5, interven_decay, interven)

        batch_embds = []
        for b in range(batch):
            features =  x[b]
            intervention = interven_adjust[b]
            for i in range(self.GNN_layers):
                features = features.contiguous().view(self.num_timestamps, self.tot_nodes, self.input_size)

                coarse_matrix = self.Coarse_module(features, intervention, adj)            # NT * NT
                fine_matrix = self.Fine_module(features, adj)                                 # NT * NT

                A_temporal = coarse_matrix * fine_matrix                                        # final ST weighted matrix
                D_temporal = Degree_Matrix(A_temporal)

                features = features.contiguous().view(-1, self.input_size)

                GCN = GConv(D_temporal, A_temporal, self.tot_nodes, self.his_temporal_weight, self.his_final_weight)
                features = GCN(features)

            his_list = []

            for timestamp in range(self.num_timestamps):
                st = timestamp * self.tot_nodes
                en = (timestamp + 1) * self.tot_nodes
                his_list.append(features[st:en, :])

            his_embds = torch.cat(his_list, dim=1)
            embds = his_embds
            embds = torch.relu(self.final_weight.mm(embds.t()).t())

            batch_embds.append(embds)

            
        batch_embds = torch.stack(batch_embds)

        out = self.output_module(batch_embds).transpose(1,2)

        return out
    
    def initialize(self):
        self.init_params()
        for layer in self.children():
            if hasattr(layer, 'reset_parameters'):
                layer.reset_parameters()






import xgboost as xgb
from sklearn.metrics import mean_squared_error
import torch.nn.init as init
import torch

class XGBModel:
    """
        Extreme Gradient Boosting (XGBoost) Model
        
        Parameters
        ----------
        num_features : int
            Number of features in each timestep of the input data.
        num_timesteps_input : int
            Number of timesteps considered for each input sample.
        num_timesteps_output : int
            Number of output timesteps to predict.
        n_estimators : int, optional
            Number of gradient boosted trees. Equivalent to the number of boosting rounds. Default: 40.
        learning_rate : float, optional
            Step size shrinkage used in update to prevents overfitting. Default: 0.1.
        max_depth : int, optional
            Maximum depth of a tree. Increasing this value will make the model more complex and more likely to overfit. Default: 5.
        reg_lambda : float, optional
            L2 regularization term on weights. Increasing this value will make model more conservative. Default: 1.0.
        reg_alpha : float, optional
            L1 regularization term on weights. Increasing this value will make model more conservative. Default: 0.1.

        Returns
        -------
        torch.Tensor
            A tensor of shape (batch_size, num_timesteps_output) representing the predicted values for the future timesteps.
            Each element corresponds to a predicted value for a future timestep.
        
    """
    def __init__(self, num_features, num_timesteps_input, num_timesteps_output,
                 n_estimators=40, learning_rate=0.1, max_depth=5, reg_lambda=1.0, reg_alpha=0.1):
        self.num_features = num_features
        self.num_timesteps_input = num_timesteps_input
        self.num_timesteps_output = num_timesteps_output
        self.n_estimators = n_estimators
        self.learning_rate = learning_rate
        self.max_depth = max_depth
        self.reg_lambda = reg_lambda
        self.reg_alpha = reg_alpha
        self.model = None

    def get_features(self, X):
        features = X.reshape(-1, self.num_features * self.num_timesteps_input)
        return features

    def fit(self, train_input, train_target, val_input=None, val_target=None, epochs=1000, batch_size=10, verbose=False, patience=100):
        """
        Parameters
        ----------
        train_input : torch.Tensor or numpy.ndarray
            The input training data, expected to be in the shape 
            (batch_size, num_timesteps_input, num_features).
        train_target : torch.Tensor or numpy.ndarray
            The target training data, expected to be in the shape
            (batch_size, num_timesteps_output).
        val_input : torch.Tensor or numpy.ndarray, optional
            The input validation data, following the same format as `train_input`.
        val_target : torch.Tensor or numpy.ndarray, optional
            The target validation data, following the same format as `train_target`.
        epochs : int, optional
            The number of epochs to train the model for. Default is 1000.
        batch_size : int, optional
            The size of batches to use when training the model. Default is 10.
        verbose : bool, optional
            If True, the training process will print out progress updates and evaluation metrics. Default is False.
        patience : int, optional
            Number of epochs with no improvement after which training will be stopped. Default is 100.

        Notes
        -----
        The method first reshapes and prepares the input data for the XGBoost regressor, converting any PyTorch tensors
        to numpy arrays if necessary. It configures the model with the specified hyperparameters and trains it on the
        provided data. If validation data is provided, it will also evaluate the model on this data after each epoch,
        which can be used for early stopping or parameter tuning. The method concludes by optionally printing the mean
        squared error of the training and validation datasets if `verbose` is True.
        """
        train_input = self.get_features(train_input)
        train_target = train_target.numpy() if isinstance(train_target, torch.Tensor) else train_target

        if val_input is not None and val_target is not None:
            val_input = self.get_features(val_input)
            val_target = val_target.numpy() if isinstance(val_target, torch.Tensor) else val_target

        self.model = xgb.XGBRegressor(
            objective='reg:squarederror', 
            n_estimators=self.n_estimators, 
            learning_rate=self.learning_rate, 
            max_depth=self.max_depth,
            reg_lambda=self.reg_lambda,  
            reg_alpha=self.reg_alpha    
        )

        if val_input is not None and val_target is not None:
            self.model.fit(train_input, train_target, eval_set=[(train_input, train_target), (val_input, val_target)], verbose=verbose)
        else:
            self.model.fit(train_input, train_target)

        # Optionally, print MSE if verbose
        if verbose:
            train_pred = self.model.predict(train_input)
            train_mse = mean_squared_error(train_target, train_pred)
            print("Training MSE: ", train_mse)
            if val_input is not None and val_target is not None:
                val_pred = self.model.predict(val_input)
                val_mse = mean_squared_error(val_target, val_pred)
                print("Validation MSE: ", val_mse)

    def predict(self, feature):
        feature = self.get_features(feature)
        prediction = self.model.predict(feature)

        return torch.tensor(prediction)



import torch
import torch.nn as nn
import torch.nn.functional as F
from torch_scatter import scatter
import numpy as np
import networkx as nx
import scipy.sparse as sp

from .base import BaseModel


class DMP(nn.Module):
    def __init__(self, 
                num_nodes,
                recover_rate=None,
                horizon=1,
                seed_list=[14,8],
                device = 'cpu'):
        super(DMP, self).__init__()

        self.n_nodes = num_nodes
        self.steps = horizon

        self.seed_list = seed_list
        self.nodes_gamma = recover_rate

    def cave_index(self, src_nodes, tar_nodes):
        edge_list = [(int(s), int(t)) for s, t in zip(src_nodes, tar_nodes)]
        E = len(edge_list)
        G = nx.DiGraph()
        G.add_edges_from(edge_list)
        attr = {edge:w for edge, w in zip(edge_list, range(E))}
        nx.set_edge_attributes(G, attr, "idx")

        cave = []
        for edge in edge_list:
            if G.has_edge(*edge[::-1]):
                cave.append(G.edges[edge[::-1]]["idx"])
            else:
                cave.append(E)
        return np.array(cave)
    
    def edgeList(self, weight_adj):
        sp_mat = sp.coo_matrix(weight_adj)
        weight = sp_mat.data
        cave = self.cave_index(sp_mat.row, sp_mat.col)
        edge_list = np.vstack((sp_mat.row, sp_mat.col, weight, cave))
        return edge_list

    def mulmul(self, Theta_t):
        Theta = scatter(Theta_t, index=self.tar_nodes, reduce="mul", dim_size=self.N) # [N]
        Theta = Theta[self.src_nodes] #[E]
        Theta_cav = scatter(Theta_t, index=self.cave_index, reduce="mul", dim_size=self.E+1)[:self.E]

        mul = Theta / Theta_cav
        return mul

    def iteration(self):
        self.Theta_ij_t = self.Theta_ij_t - self.weights * self.Phi_ij_t
        new_Ps_ij_t = self.Ps_i_0 * self.mulmul(self.Theta_ij_t)
        self.Ps_ij_t_1 = self.Ps_ij_t
        self.Ps_ij_t = new_Ps_ij_t
        self.Phi_ij_t = (1-self.weights)*(1-self.gamma)*self.Phi_ij_t - (self.Ps_ij_t-self.Ps_ij_t_1)

        # marginals
        self.Ps_t = self.Ps_0 * scatter(self.Theta_ij_t, self.tar_nodes, reduce="mul", dim_size=self.N)
        self.Pr_t = self.Pr_t + self.nodes_gamma*self.Pi_t
        self.Pi_t = 1 - self.Ps_t - self.Pr_t
        self.marginals.append([self.Ps_t, self.Pi_t, self.Pr_t])


    def _stop(self):
        I_former, R_former = self.marginals[-2][1:]
        I_later , R_later  = self.marginals[-1][1:]

        I_delta = torch.sum(torch.abs(I_former-I_later))
        R_delta = torch.sum(torch.abs(R_former-R_later))
        if I_delta>0.01 or R_delta>0.01:
            return False
        else:
            return True


    def forward(self, x, adj):
        """
        Parameters
        ----------
        x : torch.Tensor
            Expected shape (batch_size, num_timesteps_input, num_nodes, num_features).
        adj : torch.Tensor
            Adjacency matrix of the graph with shape (num_nodes, num_nodes), indicating connections between nodes.
    
        Returns
        -------
        torch.Tensor
            The output tensor of shape (horizon, num_nodes, 3), representing the predicted number of susceptible, infected, and recovered individuals at each timestep.
        """
        self.edge_list = self.edgeList(adj)
        # edge_list with size [3, E], (src_node, tar_node, weight) 
        self.src_nodes = torch.LongTensor(self.edge_list[0])
        self.tar_nodes = torch.LongTensor(self.edge_list[1])
        self.weights   = torch.FloatTensor(self.edge_list[2])
        self.cave_index = torch.LongTensor(self.edge_list[3])
        self.gamma = torch.FloatTensor(self.nodes_gamma)[self.src_nodes]
        self.nodes_gamma = torch.FloatTensor(self.nodes_gamma)
        
        self.N = max([torch.max(self.src_nodes), torch.max(self.tar_nodes)]).item()+1
        self.E = len(self.src_nodes)
        self.marginals = []

        self.seeds = torch.zeros(self.N)
        self.seeds[self.seed_list] = 1

        # initial
        self.Ps_0 = 1 - self.seeds
        self.Pi_0 = self.seeds
        self.Pr_0 = torch.zeros_like(self.seeds)

        self.Ps_i_0 = self.Ps_0[self.src_nodes]
        self.Pi_i_0 = self.Pi_0[self.src_nodes]
        self.Pr_i_0 = self.Pr_0[self.src_nodes]
        
        self.Phi_ij_0 = 1 - self.Ps_i_0
        self.Theta_ij_0 = torch.ones(self.E)      

        # first iteration, t=1
        self.Theta_ij_t = self.Theta_ij_0 - self.weights * self.Phi_ij_0 + 1E-10 # get rid of NaN
        self.Ps_ij_t_1 = self.Ps_i_0 # t-1
        self.Ps_ij_t = self.Ps_i_0 * self.mulmul(self.Theta_ij_t) # t
        self.Phi_ij_t = (1-self.weights)*(1-self.gamma)*self.Phi_ij_0 - (self.Ps_ij_t-self.Ps_ij_t_1)

        # marginals
        self.Ps_t = self.Ps_0 * scatter(self.Theta_ij_t, self.tar_nodes, reduce="mul", dim_size=self.N)
        self.Pr_t = self.Pr_0 + self.nodes_gamma*self.Pi_0
        self.Pi_t = 1 - self.Ps_t - self.Pr_t
        self.marginals.append([self.Ps_t, self.Pi_t, self.Pr_t])
        
        while True:
            self.iteration()
            if self._stop():
                break
        # Output a size of [T, N, 3] Tensor， T starts from t=1
        marginals = [torch.stack(m, dim=1) for m in self.marginals]
        marginals = torch.stack(marginals, dim=0) 


        return marginals



import torch.optim as optim
import torch.nn as nn
import torch.nn.functional as F
from copy import deepcopy
import torch
from tqdm import tqdm

from ...utils.utils import *
from ...utils.metrics import get_loss


class BaseModel(nn.Module):
    def __init__(self, device = 'cpu'):
        super(BaseModel, self).__init__()
        self.device = device

    def fit(self, 
            train_input, 
            train_target, 
            train_states=None, 
            train_graph=None, 
            train_dynamic_graph=None,
            val_input=None, 
            val_target=None,
            val_states=None, 
            val_graph= None, 
            val_dynamic_graph=None,
            loss='mse', 
            epochs=1000, 
            batch_size=10,
            lr=1e-3, 
            initialize=True, 
            verbose=False, 
            patience=100, 
            **kwargs):
        if initialize:
            self.initialize()
        
        optimizer = torch.optim.Adam(self.parameters(), lr=lr)

        loss_fn = get_loss(loss)

        training_losses = []
        validation_losses = []
        early_stopping = patience
        best_val = float('inf')
        for epoch in tqdm(range(epochs)):
            loss = self.train_epoch(optimizer=optimizer, 
                                    loss_fn=loss_fn, 
                                    feature=train_input, 
                                    states=train_states, 
                                    graph=train_graph, 
                                    dynamic_graph=train_dynamic_graph, 
                                    target=train_target, 
                                    batch_size=batch_size, 
                                    device=self.device)
            training_losses.append(loss)

            val_loss, output = self.evaluate(loss_fn=loss_fn, 
                                             feature=val_input, 
                                             graph=val_graph, 
                                             dynamic_graph=val_dynamic_graph,
                                             target=val_target, 
                                             states=val_states, 
                                             device=self.device)
            validation_losses.append(val_loss)

            if best_val > val_loss:
                best_val = val_loss
                self.best_output = output
                best_weights = deepcopy(self.state_dict())
                patience = early_stopping
            else:
                patience -= 1

            if epoch > early_stopping and patience <= 0:
                break

            if verbose and epoch%50 == 0:
                print(f"######### epoch:{epoch}")
                print("Training loss: {}".format(training_losses[-1]))
                print("Validation loss: {}".format(validation_losses[-1]))

        print("\n")
        print("Final Training loss: {}".format(training_losses[-1]))
        print("Final Validation loss: {}".format(validation_losses[-1]))
        
        self.load_state_dict(best_weights)

        
    def train_epoch(self, optimizer, loss_fn, feature, states=None, graph=None, dynamic_graph=None, target=None, batch_size=1, device='cpu'):
        """
        Trains one epoch with the given data.
        :param feature: Training features of shape (num_samples, num_nodes,
        num_timesteps_train, num_features).
        :param target: Training targets of shape (num_samples, num_nodes,
        num_timesteps_predict).
        :param batch_size: Batch size to use during training.
        :return: Average loss for this epoch.
        """
        permutation = torch.randperm(feature.shape[0])

        epoch_training_losses = []
        for i in range(0, feature.shape[0], batch_size):
            self.train()
            optimizer.zero_grad()

            indices = permutation[i:i + batch_size]
            X_batch, y_batch = feature[indices], target[indices]

            X_batch = X_batch.to(device)
            y_batch = y_batch.to(device)

            if states is not None:
                X_states = states[indices]
                X_states = X_states.to(device)
            else:
                X_states = None
            
            if dynamic_graph is not None:
                batch_graph = dynamic_graph[indices]
                batch_graph = batch_graph.to(device)
            else:
                batch_graph = None

            out = self.forward(X_batch, graph, X_states, batch_graph)
            loss = loss_fn(out, y_batch)
            loss.backward()
            optimizer.step()
            epoch_training_losses.append(loss.detach().cpu().numpy())
        return sum(epoch_training_losses)/len(epoch_training_losses)
    
    def evaluate(self, loss_fn, feature, graph = None, dynamic_graph=None, target = None, states = None, device = 'cpu'):
        with torch.no_grad():
            self.eval()
            feature = feature.to(device=device)
            target = target.to(device=device)

            out = self.forward(feature, graph, states, dynamic_graph)
            val_loss = loss_fn(out, target).to(device="cpu")
            val_loss = val_loss.detach().numpy().item()
            
            return val_loss, out

    def predict(self, feature, graph=None, states=None, dynamic_graph=None):
        """
        Returns
        -------
        torch.FloatTensor
        """
        self.eval()
        return self.forward(feature, graph, states, dynamic_graph)



import torch.nn as nn
from .base import BaseModel
import torch.nn.init as init
import torch

class GRUModel(BaseModel):
    """
        Single-layer Gated Recurrent Unit (GRU) Network

        Parameters
        ----------
        num_features : int
            Number of features in the input data.
        num_timesteps_input : int
            Number of input timesteps.
        num_timesteps_output : int
            Number of output timesteps to predict.
        nhid : int, optional
            Number of hidden units in the GRU layer. Default: 256.
        dropout : float, optional
            Dropout rate for the GRU layer. Default: 0.5.
        use_norm : bool, optional
            Whether to use Layer Normalization after the GRU layer. Default: False.

        Returns
        -------
        torch.Tensor
            A tensor of shape (batch_size, num_timesteps_output) representing the predicted values for the future timesteps.
            Each element corresponds to a predicted value for a future timestep.
            
        """
    def __init__(self, num_features, num_timesteps_input, num_timesteps_output, nhid=256, dropout=0.5, use_norm=False):
        super(GRUModel, self).__init__()
        self.num_features = num_features
        self.num_timesteps_input = num_timesteps_input
        self.num_timesteps_output = num_timesteps_output
        self.nhid = nhid
        self.dropout = dropout
        self.use_norm = use_norm

        # GRU layer
        self.gru = nn.GRU(num_features, nhid, batch_first=True)

        # Optional normalization
        if self.use_norm:
            self.norm = nn.LayerNorm(nhid)

        # Dropout layer
        self.dropout_layer = nn.Dropout(dropout)

        # Output layer
        self.out = nn.Linear(nhid, num_timesteps_output)

    def forward(self, x):
        """
        Parameters
        ----------
        x : torch.Tensor
            The input tensor for the model. Expected shape is 
            (batch_size, num_timesteps_input, num_features), where
            `batch_size` is the number of samples in the batch,
            `num_timesteps_input` is the number of input timesteps,
            and `num_features` is the number of features for each timestep.

        Returns
        -------
        torch.Tensor
            The output of the model, a tensor of shape 
            (batch_size, num_timesteps_output), representing the predicted values 
            for the future timesteps. Each element corresponds to a predicted value 
            for a future timestep.
        """

        gru_out, _ = self.gru(x)

        if self.use_norm:
            gru_out = self.norm(gru_out)

        # Apply dropout
        gru_out = self.dropout_layer(gru_out)

        # We only use the last timestep's output to predict the future series
        gru_out = gru_out[:, -1, :]

        # Output layer to predict the next steps
        output = self.out(gru_out)
        
        return output

    def initialize(self):
        # Initialize GRU weights
        for name, param in self.gru.named_parameters():
            if 'weight_ih' in name:
                init.xavier_uniform_(param.data)
            elif 'weight_hh' in name:
                init.orthogonal_(param.data)
            elif 'bias' in name:
                param.data.fill_(0)
        
        if self.use_norm:
            self.norm.reset_parameters()

        # Initialize the output layer
        init.xavier_uniform_(self.out.weight)
        self.out.bias.data.fill_(0)
